/*
Google Music Frame

Copyright 2011 Jiří Janoušek <janousek.jiri@gmail.com>

This program is free software: you can redistribute it and/or modify it 
under the terms of the GNU General Public License version 3, as published 
by the Free Software Foundation.

This program is distributed in the hope that it will be useful, but 
WITHOUT ANY WARRANTY; without even the implied warranties of 
MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR 
PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along 
with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#if GLIB_DBUS
using Fenryxo;
using Fenryxo.Logger;
//~ using Fenryxo.Config;
using Fenryxo.Path;
using Fenryxo.Widgets;

namespace GMF.MPRIS{
	public const string APPLICATION_BUS = "org.mpris.MediaPlayer2.googlemusicframe";
	
	/**
	 * MPRIS manager launches MPRIS application and MPRIS player objects
	 */
	public class Manager : GLib.Object{
		private GMF.Application app;
		private GMF.Player player;
		private Application mpris_app;
		private Player mpris_player;
		private unowned DBusConnection conn;
		
		/**
		 * Creates new manager of MPRIS interface
		 * 
		 * @param app Application object
		 * @param player player providing playback controls
		 */
		public Manager(GMF.Application app, GMF.Player player){
			this.app = app;
			this.player = player;
		}
		
		/**
		 * Starts MPRIS session
		 */
		public void start(){
			var owner_id = Bus.own_name(BusType.SESSION, APPLICATION_BUS,
				BusNameOwnerFlags.NONE,
				on_bus_acquired, on_name_acquired, on_name_lost);
			if(owner_id == 0){
				log_critical("Unable to obtain bus name %s", APPLICATION_BUS);
				var err = new ErrorDialog("Error occurred",
				"Sound Menu integration failed.");
				err.run();
			}
		}
		
		/**
		 * Register application and player MPRIS objects
		 */
		private void on_bus_acquired(DBusConnection conn, string name) {
			/* register MPRIS DBus objects */
			this.conn = conn;
			log_debug("Bus acquired: %s, registering objects", name);
			try {
				this.mpris_app = new Application(this.app);
				conn.register_object("/org/mpris/MediaPlayer2", this.mpris_app);
				this.mpris_player = new Player(this.player, conn);
				conn.register_object("/org/mpris/MediaPlayer2", this.mpris_player);
			} 
			catch(IOError e) {
				log_critical("Unable to register objects: %s", e.message);
				var err = new ErrorDialog("Error occurred",
				"Sound Menu integration failed.");
				err.run();
			}
		}
	
		private void on_name_acquired(DBusConnection connection, string name){
			log_debug("Bus name acquired: %s", name);
		}	
	
		private void on_name_lost(DBusConnection connection, string name){
			log_critical("Bus name lost: %s", name);
		}
	}
	
	
	/**
	 * MPRIS DBus wrapper for Application class.
	 * See http://www.mpris.org/2.1/spec/Root_Node.html for specification.
	 */
	[DBus(name = "org.mpris.MediaPlayer2")]
	public class Application: GLib.Object{
		private GMF.Application app;	
		/**
		 * Constructor
		 * 
		 * @param Application application object to be wrapped
		 */
		public Application(GMF.Application app){
			this.app = app;
		}
		
		/* Properties: http://www.mpris.org/2.1/spec/Root_Node.html#properties */
		[Description(nick = "Quit action availability", blurb = "If false, calling Quit will have no effect, and may raise a NotSupported error. If true, calling Quit will cause the media application to attempt to quit (although it may still be prevented from quitting by the user, for example). ")]
		public bool can_quit{
			get{return true;} 
		}
		
		[Description(nick = "Raise action availability", blurb = "If false, calling Raise will have no effect, and may raise a NotSupported error. If true, calling Raise will cause the media application to attempt to bring its user interface to the front, although it may be prevented from doing so (by the window manager, for example). ")]
		public bool can_raise{
			get{ return true; } 
		}
		
		[Description(nick = "Track list availability", blurb = "Indicates whether the /org/mpris/MediaPlayer2 object implements the org.mpris.MediaPlayer2.TrackList interface. ")]
		public bool has_track_list{
			get{ return false; }
		}
		
		[Description(nick = "Application name", blurb = "A friendly name to identify the media player to users.")]
		public string identity{
			owned get{ return "Google Music Frame"; }
		}
		
		[Description(nick = "Desktop entry name", blurb = "The basename of an installed .desktop file which complies with the Desktop entry specification, with the '.desktop' extension stripped.")]
		public string desktop_entry{ 
			owned get{ return this.app.desktop_entry; } 
		}
		
		[Description(nick = "The URI schemes supported by the media player.", blurb = "This can be viewed as protocols supported by the player in almost all cases. Almost every media player will include support for the 'file' scheme. Other common schemes are 'http' and 'rtsp'. ")]
		public string[] supported_uri_schemes{
			owned get { return {}; }
		}
		
		[Description(nick = "The mime-types supported by the media player. ", blurb = "The mime-types supported by the media player. Mime-types should be in the standard format (eg: audio/mpeg or application/ogg). ")]
		public string[] SupportedMimeTypes{
			owned get{ return {}; }
		}
		
		/* Methods: http://www.mpris.org/2.1/spec/Root_Node.html#methods */
		
		/**
		 * Brings the media player's user interface to the front using any appropriate
		 */
		public void raise(){
			try{
				this.app.activate();
			}
			catch(AppError e){
				log_error("Unable to activate instance: %s", e.message);
			}
			
		}
		
		/**
		 * Causes the media player to stop running.
		 */
		public void quit(){
			try{
				this.app.quit();
			}
			catch(AppError e){
				log_error("Unable to quit instance: %s", e.message);
			}
		}
	}
	
	
	/**
	 * MPRIS DBus object handling playback.
	 * See http://www.mpris.org/2.1/spec/Player_Node.html
	 */
	[DBus(name = "org.mpris.MediaPlayer2.Player")]
	public class Player : GLib.Object{
		private GMF.Player player;
		private DBusConnection conn;
		private HashTable<string,Variant> song_info;
		
		
		
		/**
		 * Constructs MPRIS Player
		 * 
		 * @param app Main application object
		 * @param conn active DBus connection
		 */
		public Player(GMF.Player player, DBusConnection conn){
			this.player = player;
			this.conn = conn;
			this.song_info = new HashTable<string,Variant>(null, null);

			this.notify.connect(this.send_property_change);
			this.player.playback_state_changed.connect(this.on_playback_state_changed);
			this.player.song_changed.connect(this.on_song_changed);

		}
		/**
		 * Emits PropertiesChanged signal used by the MPRIS to notify clients of changes.
		 * 
		 * @param p Param info
		 */
		private void send_property_change (ParamSpec p) {
	        var builder = new VariantBuilder (VariantType.ARRAY);
	        var invalid_builder = new VariantBuilder (new VariantType ("as"));
			Variant i;
			switch(p.name){
				case "Metadata":
					i = this.Metadata;
					builder.add ("{sv}", "Metadata", i);
					break;
				case "can-go-previous":
					i = this.can_go_previous;
					builder.add ("{sv}", "CanGoPrevious", i);
					break;
				case "can-go-next":
					i = this.can_go_next;
					builder.add ("{sv}", "CanGoNext", i);
					break;
				case "can-play":
					i = this.can_play;
					builder.add ("{sv}", "CanPlay", i);
					break;
				case "can-pause":
					i = this.can_pause;
					builder.add ("{sv}", "CanPause", i);
					break;
				case "playback-status":
					i = this.playback_status;
					builder.add ("{sv}", "PlaybackStatus", i);
					break;
				default:
					log_critical("Unhandled property: %s", p.name);
					return;
			}
			log_debug("Player: %s changed", p.name);
			var v = new Variant("(sa{sv}as)", "org.mpris.MediaPlayer2.Player", 
								builder, invalid_builder);
	        try {
	            conn.emit_signal("org.mpris.MediaPlayer2.googlemusicframe",
									"/org/mpris/MediaPlayer2", 
							"org.freedesktop.DBus.Properties", 
								"PropertiesChanged", v);
	
	        }
	        catch(Error e){
	            log_warning("Unable to emit PropertiesChanged signal: %s", e.message);
	        }
	    }
	    
	    /**
	     * Updates information passed over DBus and sends notification.
	     */
	    private void on_song_changed(string album_art, string? artist, string? song){
			var info = new HashTable<string,Variant>(null, null);
			string[] artistArray = {};
			if(artist != null) artistArray += artist;
			info.insert("xesam:artist", artistArray);
			if(song != null) info.insert("xesam:title", song);
			info.insert("mpris:artUrl", "file://" + album_art);
			this.Metadata = info;
			
		}
		
		/**
		 * Reflects changes in playback state
		 */
		private void on_playback_state_changed(string state, bool can_prev, bool can_next){
			this.playback_status = state == "none" ? "Stopped"
				: (state == "paused" ? "Paused" : "Playing");
			this.can_go_previous = can_prev;
			this.can_go_next = can_next;
			this.can_pause = this.can_play = state != "none";

		}
		
		
		/* Properties: http://www.mpris.org/2.1/spec/Player_Node.html#properties */
		
		[Description(nick = "The current playback status.", blurb = "The current playback status. May be 'Playing', 'Paused' or 'Stopped'.")]
		public string playback_status {	owned get; private set; default = "Stopped";}
		
		[Description(nick = "The current loop / repeat status", blurb = "May be: 'None' if the playback will stop when there are no more tracks to play. 'Track' if the current track will start again from the begining once it has finished playing. 'Playlist' if the playback loops through a list of tracks.")]
		public string LoopStatus{
			owned get{
				return "None";
	//~ 			return "Track";
	//~ 			return "Playlist";
			}
			set{
				switch(value){
					case("None"):
						break;
					case("Track"):
						break;
					case("Playlist"):
						break;
					default:
						break;
				}
			}
		}
	
	// Doesn't compile	
	//~ 	[Description(nick = "The current playback rate.", blurb = "The current playback rate.")]
	//~ 	public double Rate{
	//~ 		get{
	//~ 			return (double)1.0;
	//~ 		}
	//~ 		set{
	//~ 		}
	//~ 	}
	//~ 	
	//~ 	[Description(nick = "Minimum playback rate.", blurb = "Minimum playback rate.")]
	//~ 	public double MinimumRate{
	//~ 		get{
	//~ 			return (double)1.0;
	//~ 		}
	//~ 	}
	//~ 	
	//~ 	[Description(nick = "Maximum playback rate.", blurb = "Maximum playback rate.")]
	//~ 	public double MaximumRate{
	//~ 		get{
	//~ 			return (double)1.0;
	//~ 		}
	//~ 	}
		
		[Description(nick = "Shuffle state", blurb = "A value of false indicates that playback is progressing linearly through a playlist, while true means playback is progressing through a playlist in some other order.")]
		public bool Shuffle{
			get{ return true; }
			set{}
		}
		
		[Description(nick = "The metadata of the current element.", blurb = "The metadata of the current element.")]
		public HashTable<string,Variant>? Metadata{
			owned get{
				Variant variant = "1";
				this.song_info.insert("mpris:trackid", variant);
				return this.song_info;
			}
			private set{
				this.song_info = value;
			}
		}
		
	//~ 	[Description(nick = "The volume level.", blurb = "The volume level.")]
	//~ 	public double Volume{
	//~ 		owned get{
	//~ 			return (double)1.0;
	//~ 		}
	//~ 		set {
	//~ 			
	//~ 		}
	//~ 	}
		
		[Description(nick = "The current track position in microseconds.", blurb = "The current track position in microseconds.")]
		public int64 Position{
			get{
				return (int64)1;
			}
		}
		
		[Description(nick = "Next song action availability", blurb = "Whether the client can call the Next method on this interface and expect the current track to change.")]
		public bool can_go_next{ get; private set; default = false; }			
		
		[Description(nick = "Previous song action availability", blurb = "Whether the client can call the Previous method on this interface and expect the current track to change.")]
		public bool can_go_previous{ get; private set; default = false; }
		
		[Description(nick = "Play action availability", blurb = "Whether playback can be started using Play or PlayPause.")]
		public bool can_play{ get; private set; default = false; }
		
		[Description(nick = "Pause action availability", blurb = "Whether playback can be paused using Pause or PlayPause.")]
		public bool can_pause{ get; private set; default = false; }
		
		[Description(nick = "Seek action availability", blurb = "Whether the client can control the playback position using Seek and SetPosition. This may be different for different tracks.")]
		public bool CanSeek{
			get{
				return false;
			}
		}
		
		[Description(nick = "Whether the media player may be controlled.", blurb = "Whether the media player may be controlled over this interface.")]
		public bool CanControl{
			get{
				return true;
			}
		}
		
		/* Signals: http://www.mpris.org/2.1/spec/Player_Node.html#signals */
		/**
		 * Indicates that the track position has changed in a way that is inconsistant with the current playing state.
		 * @param Position The new position, in microseconds.
		 */
		public signal void Seeked(int64 Position);
		
		/* Methods: http://www.mpris.org/2.1/spec/Player_Node.html#methods */
	
		/**
		 * Skips to the next track in the tracklist.
		 * 
		 * If there is no next track (and endless playback and track repeat are both off),
		 * playback will be stopped. If playback is paused or stopped, it remains that way.
		 */
		public void next(){
			this.player.next_song(false);
		}
		
		/***
		 * Skips to the previous track in the tracklist.
		 * 
		 * If there is no previous track (and endless playback and track repeat are both off),
		 * playback will be stopped. If playback is paused or stopped, it remains that way.
		 */
		public void previous(){
			this.player.previous_song(false);
		}
		
		/**
		 * Pauses playback.
		 * 
		 * If playback is already paused, this has no effect. Calling Play after this should
		 * cause playback to start again from the same position. 
		 */
		public void pause(){
			this.player.pause(false);
		}
		
		/**
		 * Pauses playback.
		 * If playback is already paused, resumes playback.
		 * If playback is stopped, starts playback.
		 */
		public void play_pause(){
			this.player.toggle_play(false);
		}
		
		/**
		 * Stops playback.
		 * 
		 * If playback is already stopped, this has no effect. Calling Play after this should
		 * cause playback to start again from the beginning of the track. 
		 */
		public void stop(){
			this.player.stop(false);
		}
		
		/**
		 * Starts or resumes playback.
		 * 
		 * If already playing, this has no effect.
		 * If there is no track to play, this has no effect.
		 */
		public void play(){
			this.player.play(false);
		}
		
		/**
		 * Seeks forward in the current track by the specified number of microseconds.
		 * 
		 * A negative value seeks back. If this would mean seeking back further than the start
		 * of the track, the position is set to 0. If the value passed in would mean seeking
		 * beyond the end of the track, acts like a call to Next. If the CanSeek property
		 * is false, this has no effect.
		 * 
		 * @param Offset The number of microseconds to seek forward.
		 */
		public void seek(int64 Offset) {
			return;
		}
		
		/**
		 * Sets the current track position in microseconds.
		 * 
		 * If the Position argument is less than 0, does nothing. 
		 * If the Position argument is greater than the track length, does nothing.
		 * If the CanSeek property is false, this has no effect. 
		 * 
		 * @param TrackId The currently playing track's identifier. If this does not match
		 * the id of the currently-playing track, the call is ignored as "stale".
		 * @param Position Track position in microseconds.This must be between 0 and <track_length>.
		 */
		public void SetPosition(string dobj, int64 Position) {
		}
		
		/**
		 * Not implemented.
		 * @param Uri Uri of the track to load.
		 */
		public void OpenUri(string Uri) {
			return;
		}
	}

}
	
#endif
