/*
Google Music Frame

Copyright 2011 Jiří Janoušek <janousek.jiri@gmail.com>

This program is free software: you can redistribute it and/or modify it 
under the terms of the GNU General Public License version 3, as published 
by the Free Software Foundation.

This program is distributed in the hope that it will be useful, but 
WITHOUT ANY WARRANTY; without even the implied warranties of 
MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR 
PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along 
with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#if GLIB_DBUS
using Fenryxo.Logger;
namespace GMF.Helpers{
	
	/**
	 * Manager of Ubuntu Sound indicator
	 */
	public class UbuntuSoundMenu : GLib.Object{
		private GMF.Application app;
		private IndicatorSound? sound_menu;
#if INDICATOR_SOUND_MAVERICK
		// Maverick has a different registering process
		private Indicate.Server maverick_server = null;
#endif
		
		[Description(nick = "Visibility of Ubuntu Sound menu entry.", blurb = "If true a Ubuntu Sound menu entry is visible and the application is not blacklisted.")]
		public bool visible{
			get{
				
#if INDICATOR_SOUND_MAVERICK
				return this.maverick_server != null;
#else
				try{
//~ 					throw new IOError.TIMED_OUT("foo"); // Simulate IOError
					return this.sound_menu != null
					&& !this.sound_menu.is_blacklisted(this.app.desktop_entry);

				}
				catch(IOError e){
					log_critical("Ubuntu Sound Menu IOError: %s", e.message);
					// Workaround for timeout error: increase delay
					int delay = app.config.get_int("sound_indicator_delay", 1);
					log_debug(@"Ubuntu Sound Menu - current delay: $delay");
					if(delay < 5){
						delay++;
						log_debug(@"Ubuntu Sound Menu - new delay: $delay");
						this.app.config["sound_indicator_delay"] = delay.to_string();
						try{
							this.app.config.save();
						}
						catch(Error e){
							log_warning("Unable to save configuration: %s", e.message);
						}
					}
					
					return false;
				}
#endif
			}
		}
		
		/**
		 * Emmited when state of Ubuntu Sound menu is changed 
		 */
		public signal void state_changed();
		
		/**
		 * Manages Ubuntu Sound indicator for application
		 * 
		 * @param app Application object
		 */
		public UbuntuSoundMenu(GMF.Application app){
			this.app = app;
		}
		
		/**
		 * Starts watching for sound indicator
		 */
		public void start(){
			/* Watch sound menu */

#if INDICATOR_SOUND_MAVERICK
			Bus.watch_name(BusType.SESSION, "org.ayatana.indicator.sound",
				BusNameWatcherFlags.NONE,
				this.maverick_sound_menu_appeared, this.maverick_sound_menu_vanished);
#else
			Bus.watch_name(BusType.SESSION, "com.canonical.indicators.sound",
				BusNameWatcherFlags.NONE,
				this.sound_menu_appeared, this.sound_menu_vanished);
#endif
			
		}
		
		/**
		 * Creates proxy object to check whether our application is not blacklisted
		 */
		private void sound_menu_appeared(DBusConnection conn, string name, string owner){
			log_debug("Sound menu appeared: %s, %s", name, owner);
			try{
				this.sound_menu = Bus.get_proxy_sync(BusType.SESSION,
				"com.canonical.indicators.sound",
				"/com/canonical/indicators/sound/service");
				// If this.state_changed(); is called direcetly,
				// this.visible receives timeout error.
				uint delay = (uint) app.config.get_int("sound_indicator_delay", 1);
				log_debug("UbuntuSoundMenu: emmitting 'state_changed' with delay %d second(s)", delay);
				Timeout.add_seconds(delay, () => {
					this.state_changed();
					return false;
				});
			}
			catch(IOError e){
				log_warning("Unable to get proxy for sound menu: %s", e.message);
				this.sound_menu = null;
				this.state_changed();
			}
			
		}
		
		/**
		 * Removes proxy object
		 */
		private void sound_menu_vanished(DBusConnection conn, string name){
			log_debug("Sound menu vanished: %s", name);
			this.sound_menu = null;
			this.state_changed();
		}
		
#if INDICATOR_SOUND_MAVERICK
		/**
		 * Triggers registration process on Ubuntu Maverick
		 */
		private void maverick_sound_menu_appeared(DBusConnection conn, string name) {
			log_debug("Maverick sound menu appeared: %s", name);
			this.maverick_server = Indicate.Server.ref_default();
			this.maverick_server.set("type", "music." + this.app.desktop_entry);
			var entry = Fenryxo.Path.get_data_file(
				"applications/" + this.app.desktop_entry + ".desktop");
			
			if(entry == null){
				log_debug("Maverick desktop file: %s", "null");
				return;
			}
			log_debug("Maverick desktop file: %s", entry.get_path());
			this.maverick_server.set_desktop_file(entry.get_path());
			this.maverick_server.show();
			Idle.add(() => {
				this.state_changed();
				return false;
			});
		}
		
		/**
		 * Triggers unregistration process on Ubuntu Maverick
		 */
		private void maverick_sound_menu_vanished(DBusConnection conn, string name) {
			log_debug("Maverick sound menu vanished: %s", name);
			if(this.maverick_server != null){
				this.maverick_server.hide();
				this.maverick_server = null;
			}
			this.state_changed();
		}
#endif
	}


	[DBus(name = "com.canonical.indicators.sound")]
	interface IndicatorSound : Object{
		public abstract bool is_blacklisted(string player_name) throws IOError;
	}
}

#endif
